MODULE = 'SQLExecutorX'

LIMIT_1 = 1

LIMIT_2 = 2

CACHE_SIZE = 64

MYSQL_PORT, POSTGRESQL_PORT = 3306, 5432

PARAM_DRIVER, PARAM_DEBUG, PARAM_SHOW_SQL, PARAM_POOL_SIZE, PARAM_PORT = 'driver', 'debug', 'show_sql', 'pool_size', 'port'



from enum import Enum


class Engine(Enum):
    MYSQL = 'MySQL'
    POSTGRESQL = 'PostgreSQL'
    ORACLE = 'Oracle'
    SQL_SERVER = 'SQL Server'
    SQLITE = 'SQLite'
    UNKNOW = 'Unknow'


class Driver(Enum):
    MYSQL_CLIENT = 'MySQLdb'
    PYMYSQL = 'pymysql'
    MYSQL_CONNECTOR = 'mysql.connector'
    PSYCOPG2 = 'psycopg2'
    PG8000 = 'pg8000'
    PY_POSTGRESQL = 'postgresql.driver.dbapi20'
    PYGRESQL = 'pgdb'
    ORACLEDB = 'oracledb'
    SQLITE3 = 'sqlite3'


DRIVER_ENGINE_DICT = {
    Driver.MYSQL_CLIENT.value: Engine.MYSQL,
    Driver.PYMYSQL.value: Engine.MYSQL,
    Driver.MYSQL_CONNECTOR.value: Engine.MYSQL,
    Driver.PSYCOPG2.value: Engine.POSTGRESQL,
    Driver.PG8000.value: Engine.POSTGRESQL,
    Driver.PY_POSTGRESQL.value: Engine.POSTGRESQL,
    Driver.PYGRESQL.value: Engine.POSTGRESQL,
    Driver.ORACLEDB.value: Engine.ORACLE,
    Driver.SQLITE3.value: Engine.SQLITE
}


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

from typing import List, Tuple, Any
from .log_support import logger
from .engine import Engine, Driver
from .constant import MODULE, LIMIT_1, LIMIT_2
from .init_import import import_driver
from .log_support import do_sql_log, do_save_log, batch_sql_log, page_log
from .sql_support import limit_sql, require_limit, is_tuple_or_list
from .support import DBCtx, try_commit, DB_LOCK, Dict, InitArgs
from .error import MultiRowsError, MultiColumnsError
from ._context import _NoParamDecoratorContextManager

_DB_CTX = None
_POOLED = False
_SHOW_SQL = False


def init(*args, **kwargs) -> Engine:
    """
    Compliant with the Python DB API 2.0 (PEP-249).

    Addition parameters:
    :param driver=None: str|Driver, 'psycopg2' or 'pymysql' or 'mysql.connector' or 'sqlite3'
    :param pool_size=0: int, default 0, size of connection pool
    :param show_sql=False: bool,  if True, print sql
    :param debug=False: bool, if True, print debug context

    Other parameters of connection pool refer to DBUtils: https://webwareforpython.github.io/DBUtils/main.html#pooleddb-pooled-db

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> db.init('db.sqlite3', driver='sqlite3', show_sql=True debug=True)
    >>> or
    >>> db.init("postgres://user:password@127.0.0.1:5432/testdb", driver='psycopg2', pool_size=5, debug=True)
    >>> or
    >>> db.init(user='root', password='xxx', host='127.0.0.1', port=3306, database='testdb', driver='pymysql')
    """

    global _DB_CTX
    global _SHOW_SQL
    
    if _DB_CTX is not None:
        logger.warn('Database is already initialized.')
        return None
    
    pool_size = 0
    pool_args = ['mincached', 'maxcached', 'maxshared', 'maxconnections', 'blocking', 'maxusage', 'setsession', 'reset',
                 'failures', 'ping']
    
    _SHOW_SQL = kwargs.pop(InitArgs.SHOW_SQL) if InitArgs.SHOW_SQL in kwargs else False
    driver = kwargs.pop(InitArgs.DRIVER) if InitArgs.DRIVER in kwargs else None
    engine, driver_name, creator = import_driver(driver, *args, **kwargs)
    prepared = Driver.MYSQL_CONNECTOR.value == driver_name
    if InitArgs.DEBUG in kwargs and kwargs.pop(InitArgs.DEBUG):
        from logging import DEBUG
        logger.setLevel(DEBUG)

    if InitArgs.POOL_SIZE in kwargs:
        # mysql.connector 用自带连接池
        pool_size = kwargs[InitArgs.POOL_SIZE] if prepared else kwargs.pop(InitArgs.POOL_SIZE)

    pool_kwargs = {key: kwargs.pop(key) for key in pool_args if key in kwargs}
    _connect = lambda: creator.connect(*args, **kwargs)
    if pool_size >= 1 and not prepared:
        from .pooling import pooled_connect
        global _POOLED
        _POOLED = True
        _connect = pooled_connect(_connect, pool_size, **pool_kwargs)

    with DB_LOCK:
        if _DB_CTX is None:
            _DB_CTX = DBCtx(connect=_connect, prepared=prepared)
            if pool_size > 0:
                logger.info(
                    "Inited database <%s> of %s with driver: '%s' and pool size: %d." % (hex(id(_DB_CTX)), engine.value,
                                                                                         driver_name, pool_size))
            else:
                logger.info(
                    "Inited database <%s> of %s with driver: '%s'." % (hex(id(_DB_CTX)), engine.value, driver_name))
        else:
            logger.warning('Database is already initialized.')

    return engine


class conn(_NoParamDecoratorContextManager):
    r"""Context-manager that connction.

    Example::
        >>> from sqlexecutorx import conn
        >>> @conn
        >>> def foo(*args, **kw):
        >>>     f1()
        >>>     f2()
        >>>
        >>> with conn():
        >>>     pass
    """

    def __init__(self):
        super().__init__()
        self.db_ctx = _DB_CTX

    def __enter__(self):
        self.should_cleanup = False
        if self.db_ctx.try_init():
            self.should_cleanup = True
        return self

    def __exit__(self, exctype, excvalue, traceback):
        if self.should_cleanup:
            self.db_ctx.release()


class trans(_NoParamDecoratorContextManager):
    r"""Context-manager that transaction.

    Example::
        >>> from sqlexecutorx import trans
        >>> @trans
        >>> def foo(*args, **kw):
        >>>     f1()
        >>>     f2()
        >>>
        >>> with trans():
        >>>     pass
    """

    def __init__(self):
        super().__init__()
        self.db_ctx = _DB_CTX

    def __enter__(self):
        self.should_close_conn = False
        if self.db_ctx.try_init():
            # needs open a connection first:
            self.should_close_conn = True
        self.db_ctx.transactions += 1
        logger.debug('Begin transaction...' if self.db_ctx.transactions == 1 else 'Join current transaction...')
        return self

    def __exit__(self, exctype, excvalue, traceback):
        self.db_ctx.transactions -= 1
        try:
            if self.db_ctx.transactions == 0:
                if exctype is None:
                    self.commit()
                else:
                    self.rollback()
        finally:
            if self.should_close_conn:
                self.db_ctx.release()

    def commit(self):
        try_commit(self.db_ctx)

    def rollback(self):
        logger.warning('Rollback transaction...')
        self.db_ctx.connection.rollback()
        logger.debug('Rollback ok.')


def get_connection():
    _DB_CTX.try_init()
    return _DB_CTX.connection


def close():
    global _DB_CTX
    global _POOLED

    if _POOLED:
        from .pooling import close_pool
        close_pool()
        _POOLED = False

    if _DB_CTX is not None:
        _DB_CTX.release()
        _DB_CTX = None


def execute(sql: str, *args) -> int:
    """
    Execute sql return effect rowcount

    :param sql: SQL
    :param args:
    :return: Effect rowcount

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(%s, %s)'
    >>> db.execute(sql, '张三', 20)
    1
    """

    cursor = None
    should_cleanup = False
    if _SHOW_SQL:
        do_sql_log(MODULE, 'execute', sql, *args)
    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        result = cursor.rowcount
        if _DB_CTX.transactions == 0:
            try_commit(_DB_CTX)
        return result
    finally:
        if cursor:
            cursor.close()
        if should_cleanup:
            _DB_CTX.release()


def save(select_key: str, sql: str, *args) -> Any:
    """
    Execute sql return primary key

    :param select_key:
    :param sql: SQL
    :param args:
    :return: Primary key

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(%s, %s)'
    >>> db.save('SELECT LAST_INSERT_ID()', sql, '张三', 20)
    3
    """

    cursor = None
    should_cleanup = False
    if _SHOW_SQL:
        do_save_log(MODULE, 'save', select_key, sql, *args)
    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        cursor.execute(select_key)
        result = cursor.fetchone()[0]
        if _DB_CTX.transactions == 0:
            try_commit(_DB_CTX)
        return result
    finally:
        if cursor:
            cursor.close()
        if  should_cleanup:
            _DB_CTX.release()


def get(sql: str, *args) -> Any:
    """
    Execute select SQL and expected one int and only one int result, SQL contain 'limit'.
    MultiColumnsError: Expect only one column.

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT count(1) FROM person WHERE name=%s and age=%s LIMIT 1'
    >>> db.get(sql, '张三', 20)
    1
    """

    result = select_first(sql, *args)
    if result:
        if len(result) == 1:
            return result[0]
        msg = f"Exec func 'sqlexecutorx.get' expect only one column but {len(result)}."
        logger.error('%s  \n\t sql: %s \n\t args: %s' % (msg, sql, args))
        raise MultiColumnsError(msg)
    return None


def select(sql: str, *args) -> List[Tuple[Any, ...]]:
    """
    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s'
    >>> db.select(sql, '张三', 20)
    [(3, '张三', 20)]
    """
    return do_select(sql, *args)[0]


def select_first(sql: str, *args) -> Tuple[Any, ...]:
    """
    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s LIMIT 1'
    >>> db.select_first(sql, '张三', 20)
    (3, '张三', 20)
    """
    return do_select_first(sql, *args)[0]


def select_one(sql: str, *args) -> Tuple[Any, ...]:
    """
    Execute select SQL and expected one row and only one row result, SQL contain 'limit'.
    MultiRowsError: Expect only one row.

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s LIMIT 1'
    >>> db.select_one(sql, '张三', 20)
    (3, '张三', 20)
    """
    return do_select_one(sql, *args)[0]


def query(sql: str, *args) -> List[Dict[str, Any]]:
    """
    Execute select SQL and return list results(dict).

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s'
    >>> db.query(sql, '张三', 20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    """
    results, description = do_select(sql, *args)
    if results and description:
        names = list(map(lambda x: x[0], description))
        return list(map(lambda x: Dict(names, x), results))
    return results


def query_first(sql: str, *args) -> Dict[str, Any]:
    """
    Execute select SQL and return unique result(dict), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s LIMIT 1'
    >>> db.query_first(sql, '张三', 20)
    {'id': 3, 'name': '张三', 'age': 20}
    """

    result, description = do_select_first(sql, *args)
    if result and description:
        names = list(map(lambda x: x[0], description))
        return Dict(names, result)
    return result

def query_one(sql: str, *args) -> Dict[str, Any]:
    """
    Execute select SQL and return unique result(dict), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s LIMIT 1'
    >>> db.query_one(sql, '张三', 20)
    {'id': 3, 'name': '张三', 'age': 20}
    """

    result, description = do_select_one(sql, *args)
    if result and description:
        names = list(map(lambda x: x[0], description))
        return Dict(names, result)
    return result


def ravel_list(sql: str, *args, position: int = 0) -> List:
    """
    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT name FROM person WHERE age=%s'
    >>> db.ravel_list(sql, '张三', 20)
    ['张三', '李四', '王五']
    """
    return [r[position] for r in select(sql, *args)]


def do_select(sql: str, *args) -> Tuple[List[Tuple[Any, ...]], Tuple[Tuple[str, Any, Any], ...]]:
    """
    Execute select SQL and return results and description

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s limit 1'
    >>> db.do_select(sql, '张三', 20)
    ([(3, '张三', 20)], (('id', None, None), ('name', None, None), ('age', None, None)))
    """

    cursor = None
    should_cleanup = False
    if _SHOW_SQL:
        do_sql_log(MODULE, 'do_select', sql, *args)

    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        return cursor.fetchall(), cursor.description
    finally:
        if cursor:
            cursor.close()
        if should_cleanup:
            _DB_CTX.release()


def do_select_first(sql: str, *args) -> Tuple[Tuple[Any, ...], Tuple[Tuple[str, Any, Any], ...]]:
    """
    Execute select SQL and return result and description

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s limit 1'
    >>> db.do_select_first(sql, '张三', 20)
    ((3, '张三', 20), (('id', None, None), ('name', None, None), ('age', None, None)))
    """

    if require_limit(sql):
        sql = limit_sql(sql, LIMIT_1)

    return _do_select_first(sql, *args)


def _do_select_first(sql: str, *args) -> Tuple[Tuple[Any, ...], Tuple[Tuple[str, Any, Any], ...]]:
    cursor = None
    should_cleanup = False

    if _SHOW_SQL:
        do_sql_log(MODULE, '_do_select_first', sql, *args)

    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        return cursor.fetchone(), cursor.description
    finally:
        if cursor:
            cursor.close()
        if should_cleanup:
            _DB_CTX.release()


def do_select_one(sql: str, *args) -> Tuple[Tuple[Any, ...], Tuple[Tuple[str, Any, Any], ...]]:
    """
    Execute select SQL and return result and description
    MultiRowsError: Expect only one row.

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s limit 1'
    >>> db.do_select_one(sql, '张三', 20)
    ((3, '张三', 20), (('id', None, None), ('name', None, None), ('age', None, None)))
    """

    if require_limit(sql):
        sql = limit_sql(sql, LIMIT_2)
        data, description = do_select(sql, *args)
        if len(data) == 1:
            return data[0], description
        elif len(data) == 0:
            return None, description

        msg = f"Exec func 'sqlexecutorx.do_select_one' expect only one row but {len(data)}."
        logger.error('%s  \n\t sql: %s \n\t args: %s' % (msg, sql, args))
        raise MultiRowsError(msg)
    else:
        return _do_select_first(sql, *args)


def batch_execute(sql: str, *args) -> int:
    """
    Batch execute sql return effected rowcount

    :param sql: SQL to execute
    :param args: All number must have same size.
    :return: Effect rowcount

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> args = [('张三', 20), ('李四', 28)]
    >>> sql = 'INSERT INTO person(name, age) VALUES(%s, %s)'
    >>> db.batch_execute(sql, args)
    2
    >>> db.batch_execute(sql, *args)
    2
    """

    cursor = None
    should_cleanup = False
    assert args, "*args must not be empty."
    assert is_tuple_or_list(args[0]), "args must not be Tuple or List."
    
    if len(args) == 1 and len(args[0]) > 0 and is_tuple_or_list(args[0][0]):
        args = args[0]
        
    if _SHOW_SQL:
        batch_sql_log(MODULE, 'batch_execute', sql, args)

    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.executemany(sql, args)
        effect_rowcount = cursor.rowcount
        if _DB_CTX.transactions == 0:
            try_commit(_DB_CTX)
        return effect_rowcount
    finally:
        if cursor:
            cursor.close()
        if should_cleanup:
            _DB_CTX.release()
            
            
def page_select(count_sql: str, sql: str, *args) -> Tuple[int, Tuple[List, Tuple]]:
    """
    Execute select SQL and return total and list results(tuple).

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> count_sql = 'SELECT count(1) FROM person WHERE name=%s and age=%s'
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s limit %s, %s'
    >>> db.page_select(count_sql, sql, '张三', 20, 0, 10)
    (1, [(3, '张三', 20)])
    """
    return _page_select(count_sql, sql, *args)[:2]


def page_query(count_sql: str, sql: str, *args) -> Tuple[int, List[Dict[str, Any]]]:
    """
    Execute select SQL and return total and list results(dict).

    Examples
    --------
    >>> import sqlexecutorx as db
    >>> count_sql = 'SELECT count(1) FROM person WHERE name=%s and age=%s'
    >>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s limit %s, %s'
    >>> db.page_query(count_sql, sql, '张三', 20, 0, 10)
    (1, [{'id': 3, 'name': '张三', 'age': 20}])
    """
    total, results, description = _page_select(count_sql, sql, *args)
    if results and description:
        names = list(map(lambda x: x[0], description))
        return total, list(map(lambda x: Dict(names, x), results))
    return total, results
      
            
def _page_select(count_sql: str, sql: str, *args):
    cursor = None
    should_cleanup = False
    if _SHOW_SQL:
        page_log(MODULE, '_page_select', count_sql, sql, *args)
        
    count_args = args[:-2] if len(args) > 2 else ()
    
    try:
        should_cleanup = _DB_CTX.try_init()
        cursor = _DB_CTX.cursor()
        cursor.execute(count_sql, count_args)
        total = cursor.fetchone()[0]
        cursor.execute(sql, args)
        return total, cursor.fetchall(), cursor.description
    finally:
        if cursor:
            cursor.close()
        if should_cleanup:
            _DB_CTX.release()



import importlib
from typing import Union
from .log_support import logger
from .support import InitArgs
from .error import DBError
from .constant import MYSQL_PORT, POSTGRESQL_PORT
from .engine import Engine, Driver, DRIVER_ENGINE_DICT


def import_driver(driver: Union[str, Driver], *args, **kwargs):
    creator = None
    if driver:
        driver_name = driver if isinstance(driver, str) else driver.value
        if driver_name in DRIVER_ENGINE_DICT:
            engine = DRIVER_ENGINE_DICT.get(driver_name)
        else:
            engine = Driver.UNKNOWN
            logger.warning(f"Driver '{driver_name}' not support yet, may be you should adapter it by yourself.")
        creator = do_import(driver_name, engine)
    else:
        curr_engine = get_engine(*args, **kwargs)
        drivers = dict(filter(lambda x: x[1] == curr_engine, DRIVER_ENGINE_DICT.items())) if curr_engine else DRIVER_ENGINE_DICT
        for driver_name, engine in drivers.items():
            try:
                creator = importlib.import_module(driver_name)
                break
            except ModuleNotFoundError:
                pass
        if not creator:
            raise DBError(f"You may forgot install driver, may be one of {list(DRIVER_ENGINE_DICT.keys())} suit you.")
    return engine, driver_name, creator


def do_import(driver_name, curr_engine):
    try:
        return importlib.import_module(driver_name)
    except ModuleNotFoundError:
        raise DBError(f"Import {curr_engine.value} driver '{driver_name}' failed, please sure it was installed or change other driver.")


def get_engine(*args, **kwargs):
    if args and isinstance(args[0], str):
        return _get_engine_from_url(args[0])
    elif InitArgs.PORT in kwargs:
        port = kwargs[InitArgs.PORT]
        if port == MYSQL_PORT:
            return Engine.MYSQL
        elif port == POSTGRESQL_PORT:
            return Engine.POSTGRESQL
    return None


def _get_engine_from_url(url: str):
    if 'mysql://' in url:
        return Engine.MYSQL
    elif 'postgres://' in url:
        return Engine.POSTGRESQL
    elif ':memory:' == url or '://' not in url:
        return Engine.SQLITE



from sqlexecutorx.support import DB_LOCK

_POOL = None
MAX_POOL_SIZE = 32


def pooled_connect(creator, pool_size, **kwargs):
    global _POOL
    assert 1 <= pool_size <= MAX_POOL_SIZE, 'pool_size should be higher or equal to 1 and lower or equal to {}'.format(MAX_POOL_SIZE)
    if _POOL is None:
        if 'mincached' not in kwargs:
            kwargs['mincached'] = pool_size
        if 'maxcached' not in kwargs:
            kwargs['maxcached'] = pool_size
        if 'maxconnections' not in kwargs:
            kwargs['maxconnections'] = MAX_POOL_SIZE

        with DB_LOCK:
            if _POOL is None:
                try:
                    from dbutils.pooled_db import PooledDB
                except ModuleNotFoundError:
                    raise ModuleNotFoundError("No module named 'dbutils', please install it use 'pip install DBUtils'")
                _POOL = PooledDB(creator, **kwargs)
    return _POOL.connection


def close_pool():
    global _POOL
    if _POOL:
        _POOL.close()
        _POOL = None




"""
Examples
--------
>>> import sqlexecutorx as db
>>> db.init('db.sqlite3', driver='sqlite3', show_sql=True, debug=True)
Engine.SQLITE
>>> sql = 'insert into person(name, age) values(%s, %s)'
>>> db.execute(sql, '张三', 20)
1
>>> sql = 'SELECT id, name, age FROM person WHERE name=%s and age=%s'
>>> db.select(sql, '张三', 20)
[(3, '张三', 20)]
>>> db.select_first(sql, '张三', 20)
(3, '张三', 20)
>>> db.query(sql, '张三', 20)
[{'id': 3, 'name': '张三', 'age': 20}]
>>> db.query_first(sql, '张三', 20)
{'id': 3, 'name': '张三', 'age': 20}
>>> sql = 'SELECT count(1) FROM person WHERE name=%s and age=%s LIMIT 1'
>>> db.get(sql, '张三', 20)
1
"""

from .core import (
    init,
    conn,
    trans,
    get_connection,
    close,
    execute,
    save,
    get,
    select,
    select_first,
    select_one,
    do_select_one,
    query,
    query_first,
    query_one,
    ravel_list,
    do_select,
    do_select_first,
    batch_execute,
    page_select,
    page_query
)
from .engine import Engine, Driver
from .support import Dict, InitArgs
from .error import DBError


__all__ = [
    'init',
    'conn',
    'trans',
    'get_connection',
    'close',
    'execute',
    'save',
    'get',
    'select',
    'select_first',
    'query',
    'query_first',
    'do_select_first',
    'select_one',
    'query_first',
    'query_one',
    'ravel_list',
    'do_select',    
    'do_select_one',
    'batch_execute',
    'page_select',
    'page_query',
    'DBError',
    'Dict',
    'Engine',
    'Driver',
    'InitArgs'
]



from logging import basicConfig, INFO, getLogger, Formatter, StreamHandler, DEBUG

logger = getLogger(__name__)
formatter = Formatter('[%(asctime)s %(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

handler = StreamHandler()
handler.setLevel(DEBUG)
handler.setFormatter(formatter)

logger.addHandler(handler)
logger.setLevel(INFO)


def db_ctx_log(action, connection):
    logger.debug(f"{action} connection <{hex(id(connection))}>...")


def do_sql_log(module: str, function: str, sql: str, *args):
    args = args if args else ''
    logger.info(f"Exec func '{module}.{function}' \n\tsql: {sql.strip()} \n\targs: {args}")


def batch_sql_log(module: str, function: str, sql: str, args):
    args = args if args else ''
    logger.info(f"Exec func '{module}.{function}' \n\tsql: {sql.strip()} \n\targs: {args}")


def do_save_log(module: str, function: str, select_key: str, sql: str, *args):
    args = args if args else ''
    logger.info(f"Exec func '{module}.{function}', select_key: '{select_key}' \n\tsql: {sql.strip()} \n\targs: {args}")
    
    
def page_log(module: str, function: str, count_sql: str, sql: str, *args):
    logger.info(f"Exec func '{module}.{function}', \n\tcount_sql: {count_sql} \n\tsql: {sql.strip()} \n\targs: {args}")



from functools import lru_cache
from typing import List, Tuple, Union
from .conf import lru_cache_size


def is_tuple_or_list(x):
    return isinstance(x, Tuple) or isinstance(x, List)


@lru_cache(maxsize=lru_cache_size)
def require_limit(sql: str):
    lower_sql = sql.lower()
    if 'limit' not in lower_sql:
        return True
    idx = lower_sql.rindex('limit')
    if idx > 0 and ')' in lower_sql[idx:]:
        return True
    return False


@lru_cache(maxsize=lru_cache_size)
def limit_sql(sql: str, placeholder: Union[str, int] = '?'):
    return f'{sql} LIMIT {placeholder}'




import threading
from .log_support import logger, db_ctx_log

DB_LOCK = threading.RLock()


def try_commit(db_ctx):
    if db_ctx.transactions == 0:
        logger.debug('Commit transaction...')
        try:
            db_ctx.connection.commit()
            logger.debug('Commit ok.')
        except Exception:
            logger.warning('Commit failed, try rollback...')
            db_ctx.connection.rollback()
            logger.warning('Rollback ok.')
            raise
        
        
class InitArgs:
    """
    Args：
        HOST = 'host' \n
        PORT = 'port' \n
        USER = 'user' \n
        PASSWORD = 'password' \n
        DATABASE = 'database' \n
        DRIVER = 'driver' \n
        DEBUG = 'debug' \n
        SHOW_SQL = 'show_sql' \n
        POOL_SIZE = 'pool_size'
    """
    HOST = 'host'
    PORT = 'port'
    USER = 'user'
    PASSWORD = 'password'
    DATABASE = 'database'
    DRIVER = 'driver'
    DEBUG = 'debug'
    SHOW_SQL = 'show_sql'
    POOL_SIZE = 'pool_size'


class DBCtx(threading.local):
    """
    Thread local object that holds connection info.
    """

    def __init__(self, connect, prepared=False):
        self.connect = connect
        self.connection = None
        self.transactions = 0
        self.prepared = prepared

    # def is_not_init(self):
    #     return self.connection is None

    def try_init(self):
        if self.connection is None:
            self.transactions = 0
            self.connection = self.connect()
            self.log('Use')
            return True
        return False

    def release(self):
        if self.connection:
            self.log('Release')
            self.connection.close()
            self.connection = None

    def cursor(self):
        """
        Return cursor
        """
        # logger.debug('Cursor prepared: %s' % self.prepared)
        return self.connection.cursor(prepared=True) if self.prepared else self.connection.cursor()

    def log(self, action: str):
        db_ctx_log(action, self.connection)


class Dict(dict):
    def __init__(self, names=(), values=(), **kw):
        super(Dict, self).__init__(**kw)
        for k, v in zip(names, values):
            self[k] = v

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Dict' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        self[key] = value
